"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
Object.defineProperty(exports, "__esModule", { value: true });
var Core = require("@storefront/core");
var flux_capacitor_1 = require("@storefront/flux-capacitor");
exports.PADDING = 20;
exports.LOADLABEL = 'loading...';
exports.BREAKPOINT_SCROLL_DOWN = 0.75;
exports.BREAKPOINT_SCROLL_UP = 1.25;
exports.BREAKPOINT_ITEM_HEIGHT = 0.25;
exports.SCROLL_OFFSET = 50;
var InfiniteScroll = /** @class */ (function () {
    function InfiniteScroll() {
        var _this = this;
        this.searchMethods = {
            pageSize: Core.Selectors.pageSize,
            productsWithMetadata: Core.Selectors.productsWithMetadata,
            recordCount: Core.Selectors.recordCount,
            currentPage: Core.Selectors.page,
            receivePage: function (count, page) { return _this.flux.actions.receivePage(count, page); },
            fetchMore: function (state, forward) { return _this.flux.actions.fetchMoreProducts(state, forward); },
            route: flux_capacitor_1.Routes.SEARCH,
        };
        this.pastPurchaseMethods = {
            pageSize: Core.Selectors.pastPurchasePageSize,
            productsWithMetadata: Core.Selectors.pastPurchaseProductsWithMetadata,
            recordCount: Core.Selectors.pastPurchaseCurrentRecordCount,
            currentPage: Core.Selectors.pastPurchasePage,
            receivePage: function (count, page) { return _this.flux.actions.receivePastPurchasePage(count, page); },
            fetchMore: function (state, forward) { return _this.flux.actions.fetchMorePastPurchaseProducts(state, forward); },
            route: flux_capacitor_1.Routes.PAST_PURCHASE,
        };
        this.state = {
            items: [],
            lastScroll: 0,
            firstLoad: true,
            loadMore: false,
            windowScroll: false,
            isFetchingForward: false,
            isFetchingBackward: false,
            setScroll: false,
            clickMore: function () { return _this.fetchMoreItems(); },
            clickPrev: function () { return _this.fetchMoreItems(false); },
        };
        this.productTransformer = function (_a) {
            var data = _a.data, meta = _a.meta, index = _a.index;
            return (__assign({}, Core.ProductTransformer.transformer(_this.config.structure)(data), { meta: meta,
                index: index }));
        };
        this.onUpdated = function () {
            var firstItem = _this.state.items[0];
            var state = { getPage: false };
            if (firstItem) {
                if (!_this.state.loadMore) {
                    var padding = _this.calculateOffset(firstItem.index - 1);
                    _this.state.prevExists
                        ? (_this.state.wrapper.style.paddingTop = exports.PADDING + "px")
                        : (_this.state.wrapper.style.paddingTop = "0px");
                    state = __assign({}, state, { padding: padding });
                    if (_this.state.firstLoad) {
                        if (_this.state.prevExists)
                            _this.state.scroller.root.scrollTop = exports.PADDING;
                        _this.state.windowScroll
                            ? Core.utils.WINDOW().addEventListener('scroll', _this.scroll)
                            : _this.state.scroller.root.addEventListener('scroll', _this.scroll);
                    }
                }
                else if (_this.state.firstLoad) {
                    _this.state.windowScroll
                        ? Core.utils.WINDOW().addEventListener('scroll', _this.scroll)
                        : _this.state.scroller.root.addEventListener('scroll', _this.scroll);
                }
                _this.state = __assign({}, _this.state, state);
            }
            if (_this.state.setScroll)
                _this.setScroll();
        };
        this.setScroll = function () {
            var imgs = _this.state.wrapper.querySelectorAll('img') || [];
            var pageSize = _this.state.pageSize(_this.flux.store.getState());
            var count = 0;
            var imgsLoaded = function (total, size, totalImages) { return count === size || count === imgs.length; };
            if (imgs.length > 0) {
                for (var i = 0; i < Math.min(pageSize, imgs.length); i++) {
                    imgs[i].onload = function () {
                        count++;
                        if (imgsLoaded(count, pageSize, imgs.length)) {
                            _this.maintainScrollTop(_this.state.rememberScrollY);
                            _this.state.windowScroll
                                ? Core.utils.WINDOW().addEventListener('scroll', _this.scroll)
                                : _this.state.scroller.root.addEventListener('scroll', _this.scroll);
                        }
                    };
                }
                // need to still add scrollTop & scroll listener if imgs don't load
                setTimeout(function () {
                    if (!imgsLoaded(count, pageSize, imgs.length)) {
                        _this.maintainScrollTop(_this.state.rememberScrollY);
                        _this.state.windowScroll
                            ? Core.utils.WINDOW().addEventListener('scroll', _this.scroll)
                            : _this.state.scroller.root.addEventListener('scroll', _this.scroll);
                    }
                }, 500);
            }
            _this.state = __assign({}, _this.state, { setScroll: false });
        };
        this.updateProducts = function () {
            var items = _this.state.productsWithMetadata(_this.flux.store.getState()).map(_this.productTransformer);
            _this.set({
                items: items,
                setScroll: true,
                rememberScrollY: !_this.state.loadMore ? exports.PADDING : 0,
                prevExists: items.length > 0 ? items[0].index !== 1 : false,
                // tslint:disable-next-line max-line-length
                moreExists: items.length > 0 ? _this.state.recordCount(_this.flux.store.getState()) !== items[items.length - 1].index : false,
            });
            _this.state.windowScroll
                ? Core.utils.WINDOW().removeEventListener('scroll', _this.scroll)
                : _this.state.scroller.root.removeEventListener('scroll', _this.scroll);
            _this.state = __assign({}, _this.state, { elItems: _this.state.wrapper.children, firstEl: items[0], lastEl: items[items.length - 1], getPage: false, firstLoad: true });
        };
        this.setProducts = function (products) {
            var items;
            if (products.length && _this.state.items.length) {
                if (products[0].index > _this.state.items[_this.state.items.length - 1].index) {
                    items = _this.state.items.concat(products.map(_this.productTransformer));
                    _this.set({
                        items: items,
                        prevExists: items[0].index !== 1,
                        moreExists: _this.state.recordCount(_this.flux.store.getState()) !== items[items.length - 1].index,
                    });
                }
                else if (products[products.length - 1].index < _this.state.items[0].index) {
                    var pageSize = _this.state.pageSize(_this.flux.store.getState());
                    // tslint:disable-next-line max-line-length
                    var rememberScrollY = _this.calculateOffset(pageSize) +
                        (_this.state.windowScroll ? Core.utils.WINDOW().pageYOffset : _this.state.scroller.root.scrollTop);
                    items = products.map(_this.productTransformer).concat(_this.state.items);
                    _this.set(__assign({}, _this.state, { items: items, setScroll: true, rememberScrollY: rememberScrollY, prevExists: items[0].index !== 1, moreExists: _this.state.recordCount(_this.flux.store.getState()) !== items[items.length - 1].index }));
                    _this.state.windowScroll
                        ? Core.utils.WINDOW().removeEventListener('scroll', _this.scroll)
                        : _this.state.scroller.root.removeEventListener('scroll', _this.scroll);
                }
            }
            return items;
        };
        this.maintainScrollTop = function (scrollTop) { return (_this.state.scroller.root.scrollTop = scrollTop); };
        this.setFirstLoadFlag = function () { return _this.set({ firstLoad: true }); };
        this.setFetchFlags = function (_a) {
            var isFetchingForward = _a.isFetchingForward, isFetchingBackward = _a.isFetchingBackward;
            return _this.set({ isFetchingForward: isFetchingForward, isFetchingBackward: isFetchingBackward });
        };
        this.scroll = function () {
            var _a = _this.state, scroller = _a.scroller, wrapper = _a.wrapper, windowScroll = _a.windowScroll;
            var wrapperHeight = wrapper.getBoundingClientRect().height;
            var windowWrapper = Core.utils.WINDOW();
            var scrollerDOM = scroller.root;
            var scrollY = windowScroll ? windowWrapper.pageYOffset : scrollerDOM.scrollTop;
            var wrapperScrollHeight = windowScroll ? windowWrapper.document.documentElement.scrollHeight : scrollerDOM.scrollHeight;
            var wrapperClientHeight = windowScroll ? windowWrapper.innerHeight : scrollerDOM.clientHeight;
            if (_this.state.getPage) {
                _this.calculatePageChange();
            }
            // tslint:disable-next-line max-line-length
            if (!_this.state.loadMore && scrollY !== _this.state.rememberScrollY) {
                // if user is scrolling down and hits point past breakpoint, should fetch
                // tslint:disable-next-line max-line-length
                if (_this.state.lastScroll < scrollY && scrollY > wrapperScrollHeight - wrapperClientHeight - exports.SCROLL_OFFSET) {
                    // tslint:disable-next-line max-line-length
                    if (_this.state.recordCount(_this.flux.store.getState()) !== _this.state.items[_this.state.items.length - 1].index) {
                        _this.fetchMoreItems();
                    }
                    // if user is scrolling up and hits point past breakpoint, should fetch
                    // tslint:disable-next-line max-line-length
                }
                else if (_this.state.lastScroll > scrollY && scrollY < exports.SCROLL_OFFSET) {
                    if (_this.state.prevExists) {
                        _this.fetchMoreItems(false);
                    }
                }
            }
            _this.state = __assign({}, _this.state, { lastScroll: scrollY, getPage: true });
        };
        this.calculateOffset = function (totalItems) {
            var listItems = _this.state.scroller.tags['gb-list-item'];
            if (listItems && listItems.length > 0) {
                var itemDimensions = listItems[0].root.getBoundingClientRect();
                var width = _this.state.scroller.root.getBoundingClientRect().width;
                var itemsPerRow = Math.floor(width / itemDimensions.width);
                var rows = totalItems / itemsPerRow;
                return rows * itemDimensions.height;
            }
            else {
                return 0;
            }
        };
        this.calculatePageChange = function () {
            var first = _this.getItem(_this.state.firstEl.index - 1);
            var last = _this.getItem(_this.state.lastEl.index + 1);
            var scroller = _this.state.scroller.root;
            var state = _this.flux.store.getState();
            var recordCount = _this.state.recordCount(state);
            var page = _this.state.currentPage(state);
            var pageSize = _this.state.pageSize(state);
            var topCheck = _this.state.windowScroll
                ? function () { return _this.topElBelowOffsetWindow(first); }
                : function () { return _this.topElBelowOffset(first, scroller.getBoundingClientRect().top); };
            var bottomCheck = _this.state.windowScroll
                ? function () { return _this.bottomElAboveOffsetWindow(last, Core.utils.WINDOW().innerHeight); }
                : function () { return _this.bottomElAboveOffset(last, scroller.getBoundingClientRect().bottom); };
            if (first && topCheck()) {
                _this.state = __assign({}, _this.state, { firstEl: _this.state.items[_this.getIndex(_this.state.firstEl.index - pageSize)], lastEl: _this.state.items[_this.getIndex(_this.state.lastEl.index - pageSize)] });
                _this.setPage(recordCount, page - 1);
            }
            else if (last && bottomCheck()) {
                _this.state = __assign({}, _this.state, { firstEl: _this.state.items[_this.getIndex(_this.state.firstEl.index + pageSize)], lastEl: _this.state.items[_this.getIndex(Math.min(_this.state.lastEl.index + pageSize, recordCount))] });
                _this.setPage(recordCount, page + 1);
            }
        };
        this.getItem = function (recordIndex) { return _this.state.elItems[_this.getIndex(recordIndex)]; };
        this.topElBelowOffset = function (element, parentTop) {
            var _a = element.getBoundingClientRect(), top = _a.top, height = _a.height;
            return top > parentTop - height * exports.BREAKPOINT_ITEM_HEIGHT;
        };
        this.bottomElAboveOffset = function (element, parentBottom) {
            var _a = element.getBoundingClientRect(), bottom = _a.bottom, height = _a.height;
            return bottom < parentBottom + height * exports.BREAKPOINT_ITEM_HEIGHT;
        };
        this.topElBelowOffsetWindow = function (element) {
            return element.getBoundingClientRect().top > 1;
        };
        this.bottomElAboveOffsetWindow = function (element, parentBottom) {
            return element.getBoundingClientRect().top < parentBottom;
        };
        this.getIndex = function (recordIndex) { return _this.state.items.findIndex(function (item) { return item.index === recordIndex; }); };
        this.setPage = function (count, page) {
            var event;
            switch (_this.props.storeSection) {
                case Core.StoreSections.PAST_PURCHASES:
                    event = Core.Events.PAST_PURCHASE_PAGE_UPDATED;
                    break;
                case Core.StoreSections.SEARCH:
                    event = Core.Events.PAGE_UPDATED;
            }
            _this.subscribeOnce(event, _this.replaceState);
            _this.flux.store.dispatch(_this.state.receivePage(count, page));
        };
        this.replaceState = function () {
            if (!_this.state.firstLoad)
                _this.flux.replaceState(_this.state.route, true);
        };
        this.fetchMoreItems = function (forward) {
            if (forward === void 0) { forward = true; }
            _this.state = __assign({}, _this.state, { firstLoad: false });
            _this.flux.store.dispatch(_this.state.fetchMore(_this.state.pageSize(_this.flux.store.getState()), forward));
        };
    }
    InfiniteScroll.prototype.init = function () {
        switch (this.props.storeSection) {
            case Core.StoreSections.PAST_PURCHASES:
                this.state = __assign({}, this.state, this.pastPurchaseMethods, { items: this.pastPurchaseMethods.productsWithMetadata(this.flux.store.getState()).map(this.productTransformer) });
                break;
            case Core.StoreSections.SEARCH:
                this.state = __assign({}, this.state, this.searchMethods, { items: this.searchMethods.productsWithMetadata(this.flux.store.getState()).map(this.productTransformer) });
        }
    };
    InfiniteScroll.prototype.setupListeners = function () {
        switch (this.props.storeSection) {
            case Core.StoreSections.PAST_PURCHASES:
                this.subscribe(Core.Events.PAST_PURCHASE_PRODUCTS_UPDATED, this.updateProducts);
                this.subscribe(Core.Events.PAST_PURCHASE_MORE_PRODUCTS_ADDED, this.setProducts);
                this.subscribe(Core.Events.PAST_PURCHASE_CHANGED, this.setFirstLoadFlag);
                break;
            case Core.StoreSections.SEARCH:
                this.subscribe(Core.Events.PRODUCTS_UPDATED, this.updateProducts);
                this.subscribe(Core.Events.MORE_PRODUCTS_ADDED, this.setProducts);
                this.subscribe(Core.Events.SEARCH_CHANGED, this.setFirstLoadFlag);
        }
        this.subscribe(Core.Events.INFINITE_SCROLL_UPDATED, this.setFetchFlags);
    };
    InfiniteScroll.prototype.onMount = function () {
        var scroller = this.tags['gb-infinite-list'];
        var wrapper = scroller.refs.wrapper;
        var loadMore = this.props.loadMore || this.state.loadMore;
        var loaderLabel = this.props.loaderLabel || exports.LOADLABEL;
        var windowScroll = this.props.windowScroll || this.state.windowScroll;
        this.state = __assign({}, this.state, { scroller: scroller, wrapper: wrapper, loadMore: loadMore, loaderLabel: loaderLabel, windowScroll: windowScroll });
        this.setupListeners();
        this.updateProducts();
    };
    InfiniteScroll = __decorate([
        Core.configurable,
        Core.provide('infinite'),
        Core.tag('gb-infinite-scroll', require('./index.html'), require('./index.css'))
    ], InfiniteScroll);
    return InfiniteScroll;
}());
exports.default = InfiniteScroll;
//# sourceMappingURL=index.js.map