var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
import * as Core from '@storefront/core';
import { Routes } from '@storefront/flux-capacitor';
export const PADDING = 20;
export const LOADLABEL = 'loading...';
export const BREAKPOINT_SCROLL_DOWN = 0.75;
export const BREAKPOINT_SCROLL_UP = 1.25;
export const BREAKPOINT_ITEM_HEIGHT = 0.25;
export const SCROLL_OFFSET = 50;
let InfiniteScroll = class InfiniteScroll {
    constructor() {
        this.searchMethods = {
            pageSize: Core.Selectors.pageSize,
            productsWithMetadata: Core.Selectors.productsWithMetadata,
            recordCount: Core.Selectors.recordCount,
            currentPage: Core.Selectors.page,
            receivePage: (count, page) => this.flux.actions.receivePage(count, page),
            fetchMore: (state, forward) => this.flux.actions.fetchMoreProducts(state, forward),
            route: Routes.SEARCH,
        };
        this.pastPurchaseMethods = {
            pageSize: Core.Selectors.pastPurchasePageSize,
            productsWithMetadata: Core.Selectors.pastPurchaseProductsWithMetadata,
            recordCount: Core.Selectors.pastPurchaseCurrentRecordCount,
            currentPage: Core.Selectors.pastPurchasePage,
            receivePage: (count, page) => this.flux.actions.receivePastPurchasePage(count, page),
            fetchMore: (state, forward) => this.flux.actions.fetchMorePastPurchaseProducts(state, forward),
            route: Routes.PAST_PURCHASE,
        };
        this.state = {
            items: [],
            lastScroll: 0,
            firstLoad: true,
            loadMore: false,
            windowScroll: false,
            isFetchingForward: false,
            isFetchingBackward: false,
            setScroll: false,
            clickMore: () => this.fetchMoreItems(),
            clickPrev: () => this.fetchMoreItems(false),
        };
        this.productTransformer = ({ data, meta, index }) => (Object.assign({}, Core.ProductTransformer.transformer(this.config.structure)(data), { meta,
            index }));
        this.onUpdated = () => {
            const firstItem = this.state.items[0];
            let state = { getPage: false };
            if (firstItem) {
                if (!this.state.loadMore) {
                    const padding = this.calculateOffset(firstItem.index - 1);
                    this.state.prevExists
                        ? (this.state.wrapper.style.paddingTop = `${PADDING}px`)
                        : (this.state.wrapper.style.paddingTop = `0px`);
                    state = Object.assign({}, state, { padding });
                    if (this.state.firstLoad) {
                        if (this.state.prevExists)
                            this.state.scroller.root.scrollTop = PADDING;
                        this.state.windowScroll
                            ? Core.utils.WINDOW().addEventListener('scroll', this.scroll)
                            : this.state.scroller.root.addEventListener('scroll', this.scroll);
                    }
                }
                else if (this.state.firstLoad) {
                    this.state.windowScroll
                        ? Core.utils.WINDOW().addEventListener('scroll', this.scroll)
                        : this.state.scroller.root.addEventListener('scroll', this.scroll);
                }
                this.state = Object.assign({}, this.state, state);
            }
            if (this.state.setScroll)
                this.setScroll();
        };
        this.setScroll = () => {
            const imgs = this.state.wrapper.querySelectorAll('img') || [];
            const pageSize = this.state.pageSize(this.flux.store.getState());
            let count = 0;
            const imgsLoaded = (total, size, totalImages) => count === size || count === imgs.length;
            if (imgs.length > 0) {
                for (let i = 0; i < Math.min(pageSize, imgs.length); i++) {
                    imgs[i].onload = () => {
                        count++;
                        if (imgsLoaded(count, pageSize, imgs.length)) {
                            this.maintainScrollTop(this.state.rememberScrollY);
                            this.state.windowScroll
                                ? Core.utils.WINDOW().addEventListener('scroll', this.scroll)
                                : this.state.scroller.root.addEventListener('scroll', this.scroll);
                        }
                    };
                }
                // need to still add scrollTop & scroll listener if imgs don't load
                setTimeout(() => {
                    if (!imgsLoaded(count, pageSize, imgs.length)) {
                        this.maintainScrollTop(this.state.rememberScrollY);
                        this.state.windowScroll
                            ? Core.utils.WINDOW().addEventListener('scroll', this.scroll)
                            : this.state.scroller.root.addEventListener('scroll', this.scroll);
                    }
                }, 500);
            }
            this.state = Object.assign({}, this.state, { setScroll: false });
        };
        this.updateProducts = () => {
            const items = this.state.productsWithMetadata(this.flux.store.getState()).map(this.productTransformer);
            this.set({
                items,
                setScroll: true,
                rememberScrollY: !this.state.loadMore ? PADDING : 0,
                prevExists: items.length > 0 ? items[0].index !== 1 : false,
                // tslint:disable-next-line max-line-length
                moreExists: items.length > 0 ? this.state.recordCount(this.flux.store.getState()) !== items[items.length - 1].index : false,
            });
            this.state.windowScroll
                ? Core.utils.WINDOW().removeEventListener('scroll', this.scroll)
                : this.state.scroller.root.removeEventListener('scroll', this.scroll);
            this.state = Object.assign({}, this.state, { elItems: this.state.wrapper.children, firstEl: items[0], lastEl: items[items.length - 1], getPage: false, firstLoad: true });
        };
        this.setProducts = (products) => {
            let items;
            if (products.length && this.state.items.length) {
                if (products[0].index > this.state.items[this.state.items.length - 1].index) {
                    items = [...this.state.items, ...products.map(this.productTransformer)];
                    this.set({
                        items,
                        prevExists: items[0].index !== 1,
                        moreExists: this.state.recordCount(this.flux.store.getState()) !== items[items.length - 1].index,
                    });
                }
                else if (products[products.length - 1].index < this.state.items[0].index) {
                    const pageSize = this.state.pageSize(this.flux.store.getState());
                    // tslint:disable-next-line max-line-length
                    const rememberScrollY = this.calculateOffset(pageSize) +
                        (this.state.windowScroll ? Core.utils.WINDOW().pageYOffset : this.state.scroller.root.scrollTop);
                    items = [...products.map(this.productTransformer), ...this.state.items];
                    this.set(Object.assign({}, this.state, { items, setScroll: true, rememberScrollY, prevExists: items[0].index !== 1, moreExists: this.state.recordCount(this.flux.store.getState()) !== items[items.length - 1].index }));
                    this.state.windowScroll
                        ? Core.utils.WINDOW().removeEventListener('scroll', this.scroll)
                        : this.state.scroller.root.removeEventListener('scroll', this.scroll);
                }
            }
            return items;
        };
        this.maintainScrollTop = (scrollTop) => (this.state.scroller.root.scrollTop = scrollTop);
        this.setFirstLoadFlag = () => this.set({ firstLoad: true });
        this.setFetchFlags = ({ isFetchingForward, isFetchingBackward }) => this.set({ isFetchingForward, isFetchingBackward });
        this.scroll = () => {
            const { scroller, wrapper, windowScroll } = this.state;
            const wrapperHeight = wrapper.getBoundingClientRect().height;
            const windowWrapper = Core.utils.WINDOW();
            const scrollerDOM = scroller.root;
            const scrollY = windowScroll ? windowWrapper.pageYOffset : scrollerDOM.scrollTop;
            const wrapperScrollHeight = windowScroll ? windowWrapper.document.documentElement.scrollHeight : scrollerDOM.scrollHeight;
            const wrapperClientHeight = windowScroll ? windowWrapper.innerHeight : scrollerDOM.clientHeight;
            if (this.state.getPage) {
                this.calculatePageChange();
            }
            // tslint:disable-next-line max-line-length
            if (!this.state.loadMore && scrollY !== this.state.rememberScrollY) {
                // if user is scrolling down and hits point past breakpoint, should fetch
                // tslint:disable-next-line max-line-length
                if (this.state.lastScroll < scrollY && scrollY > wrapperScrollHeight - wrapperClientHeight - SCROLL_OFFSET) {
                    // tslint:disable-next-line max-line-length
                    if (this.state.recordCount(this.flux.store.getState()) !== this.state.items[this.state.items.length - 1].index) {
                        this.fetchMoreItems();
                    }
                    // if user is scrolling up and hits point past breakpoint, should fetch
                    // tslint:disable-next-line max-line-length
                }
                else if (this.state.lastScroll > scrollY && scrollY < SCROLL_OFFSET) {
                    if (this.state.prevExists) {
                        this.fetchMoreItems(false);
                    }
                }
            }
            this.state = Object.assign({}, this.state, { lastScroll: scrollY, getPage: true });
        };
        this.calculateOffset = (totalItems) => {
            const listItems = this.state.scroller.tags['gb-list-item'];
            if (listItems && listItems.length > 0) {
                const itemDimensions = listItems[0].root.getBoundingClientRect();
                const width = this.state.scroller.root.getBoundingClientRect().width;
                const itemsPerRow = Math.floor(width / itemDimensions.width);
                const rows = totalItems / itemsPerRow;
                return rows * itemDimensions.height;
            }
            else {
                return 0;
            }
        };
        this.calculatePageChange = () => {
            const first = this.getItem(this.state.firstEl.index - 1);
            const last = this.getItem(this.state.lastEl.index + 1);
            const scroller = this.state.scroller.root;
            const state = this.flux.store.getState();
            const recordCount = this.state.recordCount(state);
            const page = this.state.currentPage(state);
            const pageSize = this.state.pageSize(state);
            const topCheck = this.state.windowScroll
                ? () => this.topElBelowOffsetWindow(first)
                : () => this.topElBelowOffset(first, scroller.getBoundingClientRect().top);
            const bottomCheck = this.state.windowScroll
                ? () => this.bottomElAboveOffsetWindow(last, Core.utils.WINDOW().innerHeight)
                : () => this.bottomElAboveOffset(last, scroller.getBoundingClientRect().bottom);
            if (first && topCheck()) {
                this.state = Object.assign({}, this.state, { firstEl: this.state.items[this.getIndex(this.state.firstEl.index - pageSize)], lastEl: this.state.items[this.getIndex(this.state.lastEl.index - pageSize)] });
                this.setPage(recordCount, page - 1);
            }
            else if (last && bottomCheck()) {
                this.state = Object.assign({}, this.state, { firstEl: this.state.items[this.getIndex(this.state.firstEl.index + pageSize)], lastEl: this.state.items[this.getIndex(Math.min(this.state.lastEl.index + pageSize, recordCount))] });
                this.setPage(recordCount, page + 1);
            }
        };
        this.getItem = (recordIndex) => this.state.elItems[this.getIndex(recordIndex)];
        this.topElBelowOffset = (element, parentTop) => {
            const { top, height } = element.getBoundingClientRect();
            return top > parentTop - height * BREAKPOINT_ITEM_HEIGHT;
        };
        this.bottomElAboveOffset = (element, parentBottom) => {
            const { bottom, height } = element.getBoundingClientRect();
            return bottom < parentBottom + height * BREAKPOINT_ITEM_HEIGHT;
        };
        this.topElBelowOffsetWindow = (element) => {
            return element.getBoundingClientRect().top > 1;
        };
        this.bottomElAboveOffsetWindow = (element, parentBottom) => {
            return element.getBoundingClientRect().top < parentBottom;
        };
        this.getIndex = (recordIndex) => this.state.items.findIndex((item) => item.index === recordIndex);
        this.setPage = (count, page) => {
            let event;
            switch (this.props.storeSection) {
                case Core.StoreSections.PAST_PURCHASES:
                    event = Core.Events.PAST_PURCHASE_PAGE_UPDATED;
                    break;
                case Core.StoreSections.SEARCH:
                    event = Core.Events.PAGE_UPDATED;
            }
            this.subscribeOnce(event, this.replaceState);
            this.flux.store.dispatch(this.state.receivePage(count, page));
        };
        this.replaceState = () => {
            if (!this.state.firstLoad)
                this.flux.replaceState(this.state.route, true);
        };
        this.fetchMoreItems = (forward = true) => {
            this.state = Object.assign({}, this.state, { firstLoad: false });
            this.flux.store.dispatch(this.state.fetchMore(this.state.pageSize(this.flux.store.getState()), forward));
        };
    }
    init() {
        switch (this.props.storeSection) {
            case Core.StoreSections.PAST_PURCHASES:
                this.state = Object.assign({}, this.state, this.pastPurchaseMethods, { items: this.pastPurchaseMethods.productsWithMetadata(this.flux.store.getState()).map(this.productTransformer) });
                break;
            case Core.StoreSections.SEARCH:
                this.state = Object.assign({}, this.state, this.searchMethods, { items: this.searchMethods.productsWithMetadata(this.flux.store.getState()).map(this.productTransformer) });
        }
    }
    setupListeners() {
        switch (this.props.storeSection) {
            case Core.StoreSections.PAST_PURCHASES:
                this.subscribe(Core.Events.PAST_PURCHASE_PRODUCTS_UPDATED, this.updateProducts);
                this.subscribe(Core.Events.PAST_PURCHASE_MORE_PRODUCTS_ADDED, this.setProducts);
                this.subscribe(Core.Events.PAST_PURCHASE_CHANGED, this.setFirstLoadFlag);
                break;
            case Core.StoreSections.SEARCH:
                this.subscribe(Core.Events.PRODUCTS_UPDATED, this.updateProducts);
                this.subscribe(Core.Events.MORE_PRODUCTS_ADDED, this.setProducts);
                this.subscribe(Core.Events.SEARCH_CHANGED, this.setFirstLoadFlag);
        }
        this.subscribe(Core.Events.INFINITE_SCROLL_UPDATED, this.setFetchFlags);
    }
    onMount() {
        const scroller = this.tags['gb-infinite-list'];
        const wrapper = scroller.refs.wrapper;
        const loadMore = this.props.loadMore || this.state.loadMore;
        const loaderLabel = this.props.loaderLabel || LOADLABEL;
        const windowScroll = this.props.windowScroll || this.state.windowScroll;
        this.state = Object.assign({}, this.state, { scroller, wrapper, loadMore, loaderLabel, windowScroll });
        this.setupListeners();
        this.updateProducts();
    }
};
InfiniteScroll = __decorate([
    Core.configurable,
    Core.provide('infinite'),
    Core.tag('gb-infinite-scroll', require('./index.html'), require('./index.css'))
], InfiniteScroll);
export default InfiniteScroll;
//# sourceMappingURL=index.js.map